import torch
from torch import nn

from popgym.baselines.models.aggregations import get_aggregator


class FWPBlock(nn.Module):
    """
    The building block in the fast weight transformers paper. This is
    a form of linear transformer.

    Inputs:
        input_size: Size of input feature dim
        hidden_size: Size of key/query/value space
        aggregator: Which type of aggregation to use
        sum_normalization: Whether to use the sum normalization described
            in the paper
    """

    def __init__(
        self,
        input_size,
        hidden_size,
        aggregator="sum",
        sum_normalization=True,
    ):
        super().__init__()
        self.key = nn.Linear(input_size, hidden_size, bias=False)
        self.query = nn.Linear(input_size, hidden_size, bias=False)
        self.value = nn.Linear(input_size, hidden_size, bias=False)
        self.norm = nn.LayerNorm(input_size)
        self.sum_normalization = sum_normalization
        self.aggregator = get_aggregator(aggregator)(
            max_len=1024, d_model=hidden_size**2
        )

    def forward(self, x: torch.Tensor, state: torch.Tensor) -> torch.Tensor:
        """
        Inputs:
            x: [B, T, F]
            state: [B, 1, F]
        Outputs:
            y: [B, T, D]
            state: [B, T, F]
        """
        x = self.norm(x)
        K = self.key(x).relu()
        Q = self.query(x).relu()
        V = self.value(x)
        if self.sum_normalization:
            K = K / (1e-5 + K.sum(dim=-1, keepdim=True))
            Q = Q / (1e-5 + Q.sum(dim=-1, keepdim=True))

        kv = torch.einsum("bti, btj -> btij", V, K)
        shape = kv.shape
        state = self.aggregator(kv.flatten(-2), state.flatten(-2)).reshape(shape)

        y = torch.einsum("btij, bti -> btj", state, Q)
        return y, state



from typing import Any, Dict, Optional, Tuple

import gymnasium as gym
import numpy as np
from gymnasium.core import ActType, ObsType

from popgym.core.deck import Deck
from popgym.core.env import POPGymEnv


class RepeatPrevious(POPGymEnv):
    """A game where the agent must repeat the suit of the k-previous card it saw

    Args:
        num_decks: The number of decks to cycle through, which determines
            episode length
        k: The "previous" timestep. k == 0 denotes the current timestep, k==1
            denotes the previous timestep, ... k == 16 denotes 15 timesteps ago

    Returns:
        A gym environment
    """

    def __init__(self, num_decks=1, k=4):
        self.deck = Deck(num_decks)
        self.k = k
        self.max_episode_length = self.deck.num_cards - 1
        assert self.deck.num_cards > k, "k cannot be less than 52 * num_decks"
        self.deck.add_players("player")
        self.action_space = self.deck.get_obs_space(["suits"])
        self.observation_space = self.action_space
        self.state_space = gym.spaces.Tuple(
            (gym.spaces.MultiDiscrete([4] * k), gym.spaces.Box(0, 1, (4,)))
        )
        self.dealt_cards = np.zeros((4,), dtype=int)

    def get_state(self):
        cards = self.deck.show("player", ["suits_idx"])[0, -self.k :]
        if len(cards) != self.k:
            cards_ = np.zeros((self.k,), dtype=cards.dtype)
            cards_[-len(cards) :] = cards
            cards = cards_

        dealt_cards = 1.0 - self.dealt_cards / (self.deck.num_cards / 4)
        dealt_cards = dealt_cards.astype(np.float32)

        return cards.copy(), dealt_cards

    def step(self, action: ActType) -> Tuple[ObsType, float, bool, bool, dict]:
        reward_scale = 1 / (self.deck.num_cards - self.k)
        reward = 0

        truncated = len(self.deck) == 1

        if self.deck.hand_size("player") >= self.k:
            if action == self.deck.suits_idx[self.deck["player"][-self.k]]:
                reward = reward_scale
            else:
                reward = -reward_scale

        self.deck.deal("player", 1)
        card = self.deck.show("player", ["suits_idx"])[0, -1]
        self.dealt_cards[card] += 1
        obs = card.item()

        info: dict = {}

        return obs, reward, False, truncated, info

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict] = None,
    ) -> Tuple[gym.core.ObsType, Dict[str, Any]]:
        super().reset(seed=seed)
        self.deck.reset(rng=self.np_random)
        self.deck.deal("player", 1)
        self.card = self.deck.show("player", ["suits_idx"])[0, -1]
        self.dealt_cards[:] = 0
        self.dealt_cards[self.card] += 1
        obs = self.card.item()
        return obs, {}


class RepeatPreviousEasy(RepeatPrevious):
    pass


class RepeatPreviousMedium(RepeatPrevious):
    def __init__(self):
        super().__init__(num_decks=2, k=32)


class RepeatPreviousHard(RepeatPrevious):
    def __init__(self):
        super().__init__(num_decks=3, k=64)



import numpy as np

from popgym.envs.concentration import Concentration
from tests.base_env_test import AbstractTest


class TestConcentration(AbstractTest.POPGymTest):
    def setUp(self) -> None:
        self.env = Concentration()

    def test_repeatedly_flip_faceup(self):
        # Ensure we cannot match facedown cards
        # with previously-matched faceup cards
        obs, _ = self.env.reset()
        r_c = 0
        obs, r, te, tr, _ = self.env.step(0)
        r_c += r
        cards = [obs[0]]
        found = False
        i = 0
        while not (te or tr) and not found:
            i += 1
            obs, r, te, tr, _ = self.env.step(i)
            r_c += r
            found = obs[i] in cards
            cards.append(obs[i])
        j = cards.index(obs[i])
        if not (te or tr) and i % 2 != 1:
            obs, r, te, tr, _ = self.env.step(j)
            r_c += r
        while not (te or tr):
            obs, r, te, tr, _ = self.env.step(i)
            r_c += r
            i, j = j, i

        self.assertAlmostEqual(
            r_c,
            self.env.success_reward_scale
            + (self.env.episode_length - 2) * self.env.failure_reward_scale,
        )

    def test_perfect_game(self):
        obs, _ = self.env.reset()
        ranks = self.env.deck.ranks_idx[self.env.deck.idx]
        done = False
        running_reward = 0
        for i in range(13):
            idx = np.where(ranks == i)[0]
            for j in range(4):
                self.assertFalse(done)
                obs, reward, terminated, truncated, info = self.env.step(idx[j])
                self.assertLessEqual(len(self.env.deck["in_play"]), 2)
                running_reward += reward
                if j % 2 == 1:
                    self.assertEqual(reward, self.env.success_reward_scale)

        self.assertTrue(terminated)
        self.assertFalse(truncated)
        self.assertAlmostEqual(running_reward, 1.0)

    def test_two_decks(self):
        e = Concentration(num_decks=2)
        obs, _ = e.reset()
        ranks = e.deck.ranks_idx[e.deck.idx]
        terminated = truncated = False
        running_reward = 0
        for i in range(13):
            idx = np.where(ranks == i)[0]
            for j in range(8):
                obs, reward, terminated, truncated, info = e.step(idx[j])
                running_reward += reward
                if j % 2 == 1:
                    self.assertEqual(reward, e.success_reward_scale)

        self.assertTrue(terminated)
        self.assertFalse(truncated)
        self.assertAlmostEqual(running_reward, 1.0)

    def test_perfect_game_colors(self):
        e = Concentration(deck_type="colors")
        obs, _ = e.reset()
        colors = e.deck.colors_idx[e.deck.idx]
        terminated = truncated = False
        running_reward = 0
        for i in range(2):
            idx = np.where(colors == i)[0]
            for j in range(26):
                obs, reward, terminated, truncated, info = e.step(idx[j])
                running_reward += reward
                if j % 2 == 1:
                    print(e.deck["in_play"])
                    self.assertEqual(reward, e.success_reward_scale)

        self.assertTrue(terminated)
        self.assertFalse(truncated)
        self.assertAlmostEqual(running_reward, 1.0)

    def test_worst_game(self):
        obs, _ = self.env.reset()
        terminated = truncated = False
        running_reward = 0
        while not (terminated or truncated):
            obs, reward, terminated, truncated, info = self.env.step(0)
            running_reward += reward

        self.assertAlmostEqual(running_reward, -1.0)
        self.assertTrue(truncated)
        self.assertFalse(terminated)

    def test_worst_game_colors(self):
        e = Concentration(deck_type="colors")
        obs, _ = e.reset()
        terminated = truncated = False
        running_reward = 0
        while not (terminated or truncated):
            obs, reward, terminated, truncated, info = e.step(0)
            running_reward += reward

        self.assertAlmostEqual(running_reward, -1.0)
        self.assertTrue(truncated)
        self.assertFalse(terminated)



from popgym.envs.repeat_first import RepeatFirst
from tests.base_env_test import AbstractTest


class TestRepeatFirst(AbstractTest.POPGymTest):
    def setUp(self) -> None:
        self.env = RepeatFirst()

    def test_perfect(self):
        terminated = truncated = False
        init_item, _ = self.env.reset()
        reward = 0
        for i in range(51):
            self.assertFalse(terminated or truncated)
            item, rew, terminated, truncated, info = self.env.step(init_item)
            self.assertTrue(item < 4)
            reward += rew

        self.assertTrue(truncated)
        self.assertFalse(terminated)
        self.assertAlmostEqual(1.0, reward)



import math
from typing import List, Tuple

import gym
import torch
from ray.rllib.utils.typing import ModelConfigDict, TensorType
from torch import nn

from popgym.baselines.models.linear_attention import LinearAttentionBlock
from popgym.baselines.ray_models.base_model import BaseModel


class LinearAttention(BaseModel):
    r"""The Fast Autoregressive Transformer (FART, lol) from

    @inproceedings{katharopoulos_transformers_2020,
        title = {
            Transformers are RNNs: Fast Autoregressive
            Transformers with Linear Attention
        },
        shorttitle = {Transformers are {RNNs}},
        url = {https://proceedings.mlr.press/v119/katharopoulos20a.html},
        language = {en},
        urldate = {2022-09-21},
        booktitle = {
            Proceedings of the 37th {International}
            {Conference} on {Machine} {Learning}
        },
        publisher = {PMLR},
        author = {
            Katharopoulos, Angelos and Vyas, Apoorv and Pappas,
            Nikolaos and Fleuret, François
        },
        month = nov,
        year = {2020},
        note = {ISSN: 2640-3498},
        pages = {5156--5165},
    }
    """

    MODEL_CONFIG = {
        # Which positional embedding to use
        "embedding": "sine",
        # Which aggregation operator to use for the S term.
        # The paper only uses sum, but this can be sum or max
        "S_aggregator": "sum",
        # Same as S aggregator, except for the Z (denominator) term
        "Z_aggregator": "sum",
        # Whether the recurrent state size should be hidden_size
        # or hidden_size ** 2
        "big_hidden": False,
    }

    def __init__(
        self,
        obs_space: gym.spaces.Space,
        action_space: gym.spaces.Space,
        num_outputs: int,
        model_config: ModelConfigDict,
        name: str,
        **custom_model_kwargs,
    ):
        super().__init__(obs_space, action_space, num_outputs, model_config, name)
        if self.cfg["big_hidden"]:
            self.h = self.cfg["hidden_size"]
        else:
            self.h = round(math.sqrt(self.cfg["hidden_size"]))
        self.core = LinearAttentionBlock(
            input_size=self.cfg["preprocessor_output_size"],
            hidden_size=self.h,
            S_aggregator=self.cfg["S_aggregator"],
            Z_aggregator=self.cfg["Z_aggregator"],
        )
        if not self.cfg["big_hidden"]:
            self.unmap = nn.Linear(self.h, self.cfg["hidden_size"])

    def initial_state(self) -> List[TensorType]:
        return [
            torch.zeros(1, self.h, self.h),
            torch.zeros(1, self.h),
        ]

    def forward_memory(
        self,
        z: TensorType,
        state: List[TensorType],
        t_starts: TensorType,
        seq_lens: TensorType,
    ) -> Tuple[TensorType, List[TensorType]]:
        B, T, _ = z.shape
        z, state = self.core(z, state)
        h = z.shape[-1]
        if not self.cfg["big_hidden"]:
            z = self.unmap(z)
        S, Z = state
        return z, [
            S[:, -1].reshape(B, 1, h, h),
            Z[:, -1].reshape(B, 1, h),
        ]


class BigLinearAttention(LinearAttention):
    MODEL_CONFIG = {
        # Which positional embedding to use
        "embedding": "sine",
        # Which aggregation operator to use for the S term.
        # The paper only uses sum, but this can be sum or max
        "S_aggregator": "sum",
        # Same as S aggregator, except for the Z (denominator) term
        "Z_aggregator": "sum",
        # Whether the recurrent state size should be hidden_size
        # or hidden_size ** 2
        "big_hidden": True,
    }



import pytest
from gymnasium.utils.env_checker import check_env

from popgym import ALL_ENVS


class TestEnvs:
    @pytest.mark.parametrize("env", ALL_ENVS)
    def test_no_warn(self, env):
        e = env()
        e.reset()
        check_env(e, skip_render_check=True)



import os
import random
from multiprocessing import Manager, Pool
from typing import Any

import gym
import numpy as np
import tqdm
from ray.rllib.evaluation.sample_batch_builder import SampleBatchBuilder
from ray.rllib.models.preprocessors import get_preprocessor
from ray.rllib.offline.json_writer import JsonWriter

import popgym


class RandomPolicy:
    """A policy that randomly samples actions"""

    def __init__(self, env: gym.Env):
        self.space = env.action_space
        # Cache for faster inference
        self.actions = [self.space.sample() for i in range(10_000)]

    def __call__(self, observation: np.ndarray) -> Any:
        return random.choice(self.actions)


def generate_random_datasets(
    directory: str,
    num_timesteps: int = 10_000_000,
    num_workers: int = 8,
    max_filesize: int = 2 * 1024 * 1024 * 1024,  # 2 GiB
) -> None:
    pool = Pool(num_workers)
    m = Manager()
    q = m.Queue()
    for i in range(1, num_workers + 1):
        q.put(i)
    args = []
    for env, info in popgym.ALL_ENVS.items():
        subdir = info["id"]
        path = f"{directory}/random/{subdir}"
        os.makedirs(path, exist_ok=True)
        instance = env()
        print(f"Generating dataset for {env.__name__}...")
        args.append(
            [path, instance, RandomPolicy(instance), num_timesteps, max_filesize, q]
        )

    pool.starmap(generate_dataset, args)


def generate_dataset(
    directory: str,
    env: gym.Env,
    policy: Any,
    num_timesteps: int,
    max_filesize=2 * 1024 * 1024 * 1024,  # 2 GiB
    queue=None,
    keep_info=False,
) -> None:
    """Generate datasets using a specified policy, exporting datasets to RLlib's
    SampleBatch format as json."""
    builder = SampleBatchBuilder()
    writer = JsonWriter(directory, max_file_size=max_filesize)
    preprocessor = get_preprocessor(env.observation_space)(env.observation_space)
    transition_id = 0
    last_progress_update = 0

    worker_id = queue.get(timeout=10) if queue is not None else 0

    ep_id = 0
    progress = tqdm.tqdm(
        desc=type(env).__name__,
        unit="step",
        position=worker_id,
        total=num_timesteps,
        bar_format="{desc:<30}{percentage:3.0f}%|{bar:10}{r_bar}",
    )
    while True:
        t = 0
        done = False
        obs = env.reset(seed=ep_id, return_info=False)
        prev_action = np.zeros_like(env.action_space.sample())
        prev_reward = 0
        transition_id += 1

        while not done:
            action = policy(obs)
            new_obs, reward, done, info = env.step(action)
            builder.add_values(
                t=t,
                eps_id=ep_id,
                agent_index=0,
                obs=preprocessor.transform(obs),
                action_prob=1.0,
                action_logp=0,
                actions=action,
                rewards=reward,
                prev_action=prev_action,
                prev_rewards=prev_reward,
                dones=done,
                infos=info if keep_info else {},
                new_obs=preprocessor.transform(new_obs),
                transition_id=transition_id,
            )

            obs = new_obs
            prev_action = action
            prev_reward = reward
            t += 1
            transition_id += 1

        ep_id += 1

        if transition_id - last_progress_update > 10_000:
            progress.update(transition_id - last_progress_update)
            last_progress_update = transition_id

        if transition_id > num_timesteps:
            queue.put(worker_id)
            return

        writer.write(builder.build_and_reset())


if __name__ == "__main__":
    generate_random_datasets("/home/popgym/popgym/datasets")



from typing import Any, Dict, List, Tuple

import gym
import torch
from ray.rllib.models.torch.torch_modelv2 import TorchModelV2
from ray.rllib.utils.typing import ModelConfigDict, TensorType
from torch import nn

from popgym.baselines.models.embeddings import (
    Base2Embedding,
    PerceptronEmbedding,
    ScaledPositionalEncoding,
    SimpleScaledEncoding,
    Time2Vec,
    positional_encoding,
)


def weight_init(m):
    if isinstance(m, torch.nn.Linear):
        torch.nn.init.orthogonal_(m.weight)


class BaseModel(TorchModelV2, nn.Module):
    """The base memory model that all other memory models should
    inherit from.

    You will need to override initial_state and
    memory forward, as well as set your model defaults using
    MODEL_CONFIG.
    """

    BASE_CONFIG = {
        # Number of weights per controller hidden layer
        "hidden_size": 256,
        # Observation goes through this torch.nn.Module before
        # feeding to memory
        "preprocessor": torch.nn.Identity(),
        "postprocessor": torch.nn.Identity(),
        "actor": nn.Sequential(nn.Linear(256, 256), nn.LeakyReLU(inplace=True)),
        "critic": nn.Sequential(nn.Linear(256, 256), nn.LeakyReLU(inplace=True)),
        # The output size of the preprocessor
        "preprocessor_input_size": 256,
        # Input size to the preprocessor
        "preprocessor_output_size": 256,
        "postprocessor_output_size": 256,
        # Either none, learned, absolute, or relative
        "embedding": None,
        # Dims to use for embedding
        # if set to None, uses the entire feature vector
        "embedding_size": None,
        # Either add or concatenate the embedding
        "embedding_mode": "add",
        # Whether to learn the embedding scale
        "scale_embedding": True,
    }
    # Override me in subclasses
    MODEL_CONFIG: Dict[str, Any] = {}

    def __init__(
        self,
        obs_space: gym.spaces.Space,
        action_space: gym.spaces.Space,
        num_outputs: int,
        model_config: ModelConfigDict,
        name: str,
        **custom_model_kwargs,
    ):
        nn.Module.__init__(self)
        super().__init__(obs_space, action_space, num_outputs, model_config, name)
        self.num_outputs = num_outputs
        self.obs_dim = gym.spaces.utils.flatdim(obs_space)
        self.act_space = action_space
        self.act_dim = gym.spaces.utils.flatdim(action_space)

        self.input_dim = self.obs_dim
        for k in custom_model_kwargs:
            assert (
                k in self.BASE_CONFIG or k in self.MODEL_CONFIG
            ), f"Invalid config arg {k}"
        self.cfg = {
            **self.BASE_CONFIG,
            **self.MODEL_CONFIG,
            **custom_model_kwargs,
            **model_config.get("custom_model_config", {}),
        }
        self.feature_map = torch.nn.Sequential(
            torch.nn.Linear(self.input_dim, self.cfg["preprocessor_input_size"]),
            torch.nn.LayerNorm(
                self.cfg["preprocessor_input_size"], elementwise_affine=False
            ),
        )
        # When rollout_mode != complete_episodes, we can have significantly
        # longer rollout fragments than max_seq_len due to additional zero-padding.
        # To fix, pass 2 * max_seq_len to setup_embedding to account for this
        # but this uses more memory and results in lower model accuracy...
        self.setup_embedding(model_config["max_seq_len"])
        self.max_seq_len = model_config["max_seq_len"]
        if self.cfg["scale_embedding"]:
            self.embedding_strength = nn.Parameter(torch.tensor([0.5]))
        self.pre = self.cfg["preprocessor"]
        self.core = nn.Identity()
        self.post = self.cfg["postprocessor"]

        actor_final = nn.Linear(self.cfg["postprocessor_output_size"], self.num_outputs)
        critic_final = nn.Linear(self.cfg["postprocessor_output_size"], 1)

        self.feature_map.apply(weight_init)
        self.pre.apply(weight_init)
        self.post.apply(weight_init)
        self.cfg["actor"].apply(weight_init)
        self.cfg["critic"].apply(weight_init)
        nn.init.normal_(actor_final.weight, std=0.01)
        nn.init.normal_(critic_final.weight, std=0.01)

        self.actor = nn.Sequential(self.cfg["actor"], actor_final)
        self.critic = nn.Sequential(self.cfg["critic"], critic_final)

        self.features = None

    def setup_embedding(self, max_seq_len: int):
        assert self.cfg["embedding_mode"] in [
            "cat",
            "add",
        ], "Embedding mode must be 'add' or 'cat'"
        if self.cfg["embedding_mode"] == "cat":
            assert (
                self.cfg["embedding_size"] is not None
            ), "Must specify embedding size with 'cat' embedding mode"
        size = self.cfg["embedding_size"] or self.cfg["preprocessor_output_size"]
        if self.cfg["embedding"] == "learned":
            self.embedding = nn.Embedding(max_seq_len, size)
        elif self.cfg["embedding"] == "fixed":
            self.embedding = nn.Embedding(max_seq_len, size)
            self.embedding.requires_grad = False
            self.embedding.weight.requires_grad = False
        elif self.cfg["embedding"] == "sine":
            self.embedding = nn.Embedding.from_pretrained(
                positional_encoding(max_seq_len, size)
            )
        elif self.cfg["embedding"] == "scaled_sine":
            self.embedding = ScaledPositionalEncoding.from_pretrained(
                positional_encoding(max_seq_len, size)
            )
        elif self.cfg["embedding"] == "simple_scaled":
            self.embedding = SimpleScaledEncoding(max_seq_len, size)
        elif self.cfg["embedding"] == "perceptron":
            self.embedding = PerceptronEmbedding(max_seq_len, size)
        elif self.cfg["embedding"] == "base2":
            self.embedding = Base2Embedding(max_seq_len, size)
        elif self.cfg["embedding"] == "time2vec":
            self.embedding = Time2Vec(max_seq_len, size)
        elif self.cfg["embedding"] is None:
            self.embedding = None
        else:
            raise NotImplementedError()

    def initial_state(self) -> List[TensorType]:
        """Return the initial states for your memory model.
        The shape of states returned here should NOT contain
        the batch dimension. The batch dimension will be prepended
        by RLlib, depending on the number of episodes/rollouts
        in the batch

        Returns:
            List of tensors denoting the zeroth recurrent state"""

        raise NotImplementedError()

    def get_initial_state(self) -> List[TensorType]:
        return self.initial_state() + [
            torch.tensor([0], dtype=torch.long),
        ]

    def value_function(self) -> TensorType:
        assert self.features is not None, "must call forward() first"
        value = self.critic(self.features).reshape(-1)
        return value

    def preprocess_obs(self, obs, seq_lens):
        """Preprocess the incoming observation of shape [?, Feature] into
        [Batch, Time, Feature]
        """
        if self.view_requirements["obs"].shift == 0:
            # Recurrent models do not need to shift the observations
            B = seq_lens.shape[0]
            orig_T = obs.shape[0] // B
            unflat = obs.reshape(B, orig_T, *self.view_requirements["obs"].space.shape)
            # Sometimes RLlib sticks on extra padding for no reason
            # Remove that padding
            T = seq_lens.max().item()
            unflat = unflat[:, :T]

        else:
            # We are in attention-mode and will always receive
            # obs of shape [B, max_seq_len, F]
            B = seq_lens.shape[0]
            T = len(self.view_requirements["obs"].shift_arr)
            orig_T = T
            unflat = obs.reshape(B, T, *self.view_requirements["obs"].space.shape)

        return unflat, B, T, orig_T

    def forward_memory(
        self,
        z: TensorType,
        state: List[TensorType],
        t_starts: TensorType,
        seq_lens: TensorType,
    ) -> Tuple[TensorType, List[TensorType]]:
        """Forward for your custom memory model.

        Inputs:
            z: Preprocessed features of shape [B, T, F], with padding along the time
                dimension
            state: Recurrent states of shape [B, ...]
            t_starts: Tensor of size [B] denoteint the length of the rollout so far.
                Note that in some cases RLlib might chunk a long rollout into multiple
                forward passes. This tracks the length throughout all forward passes.
            seq_lens: Tensor of size [B] denoting the number of non-padding elements
                in z. E.g. seq_lens == [100, 60], z.shape[1] == 128 means the first 100
                elements of the first batch dimension are valid and the first 60
                elements of the second batch dimension are valid. Rest are padding.

        Returns:
            output: Output features of shape [B, T, D]. Note that the padding
                must be present here, don't worry we won't actually use it.
            state: Output states of shape [B, ...]. Note that these shapes should
                be exactly the same as input state shapes or you will have issues.
        """

        raise NotImplementedError()

    def apply_embedding(self, x, t_starts):
        """Applies temporal embeddings to x"""
        B, T, F = x.shape
        # indices used to include padding
        # note that T could be > seq_lens.max() due to shoddy RLlib batch-splitting
        time_idx = t_starts.reshape(-1, 1) + torch.arange(T, device=x.device).reshape(
            1, -1
        )

        emb = self.embedding(time_idx).reshape(B, T, -1)

        if self.cfg["scale_embedding"]:
            self.embedding_strength.data.clamp_(0.0, 1.0)
            x = x * (1.0 - self.embedding_strength)
            emb = emb * self.embedding_strength

        if self.cfg["embedding_mode"] == "add":
            if self.cfg["embedding_size"] is None:
                x = x + emb
            else:
                x[:, :, : self.cfg["embedding_size"]] = (
                    x[:, :, : self.cfg["embedding_size"]] + emb
                )
        else:
            x[:, :, : self.cfg["embedding_size"]] = emb

        return x

    def forward(
        self,
        input_dict: Dict[str, TensorType],
        state: List[TensorType],
        seq_lens: TensorType,
    ) -> Tuple[TensorType, List[TensorType]]:
        # Sometimes int64, sometimes int32
        seq_lens = seq_lens.long()
        # This is always float, fucking rllib doesn't respect dtype
        state[-1] = state[-1].long()
        x, B, T, orig_t = self.preprocess_obs(input_dict["obs_flat"], seq_lens)
        # Bug, RLlib will sometimes fail to set self.train() during training
        # and LSTM fails to backprop when self.train() not set
        self.train() if torch.is_grad_enabled() else self.eval()

        # Start and end t indices into the padded sequence
        # denoting non-padded elements
        t_starts = state[-1].squeeze(1)
        t_ends = t_starts + seq_lens

        x = self.feature_map(x)
        if self.cfg["embedding"] is not None:
            x = self.apply_embedding(x, t_starts)
        x = self.pre(x)
        x, state = self.forward_memory(x, state[:-1], t_starts, seq_lens)
        assert x.shape == (B, T, self.cfg["hidden_size"]), (
            f"Expected memory to output shape {(B, T, self.cfg['hidden_size'])} but got"
            f" {x.shape}"
        )

        features = self.post(x)
        # Add the extraneous zero-padding back (RLlib needs it)
        self.features = torch.cat(
            [
                features,
                torch.zeros(B, orig_t - T, features.shape[-1], device=x.device),
            ],
            dim=1,
        )
        logits = self.actor(self.features).reshape(B * orig_t, self.num_outputs)

        return logits, state + [t_ends.unsqueeze(1)]



import unittest

import numpy as np

from popgym.envs.bipedal_walker import BipedalWalker


class TestRepeatFirst(unittest.TestCase):
    def test_all(self):
        e = BipedalWalker()
        _ = e.reset()
        for i in range(100):
            _, _, done, _ = e.step(np.array([0.0, 0.0, 0.0, 0.0]))
            if done:
                e.reset()



from popgym.wrappers.antialias import Antialias
from popgym.wrappers.markovian import Markovian
from popgym.wrappers.previous_action import PreviousAction

__all__ = ["Antialias", "Markovian", "PreviousAction"]



import unittest

from popgym.envs.labyrinth_escape import LabyrinthEscape


class TestLabyrinthEscape(unittest.TestCase):
    def test_step(self):
        e = LabyrinthEscape()
        e.reset()
        for i in range(1000):
            _, _, terminated, truncated, _ = e.step(e.action_space.sample())
            if terminated or truncated:
                e.reset()

    def test_tostring(self):
        e = LabyrinthEscape((6, 6))
        e.reset()
        e.tostring()

    def test_goal(self):
        e = LabyrinthEscape((6, 6))
        terminated = truncated = False
        e.reset()
        for i in range(5):
            while not (terminated or truncated):
                _, _, terminated, truncated, _ = e.step(e.action_space.sample())
            self.assertTrue(e.curr_step < e.max_episode_length)
            e.reset()

    def test_known(self):
        e = LabyrinthEscape((8, 8))
        e.reset(seed=1)
        left, right, _, down = 0, 1, 2, 3
        actions = [down] * 5 + [left] * 2 + [down] * 2 + [right] * 6 + [down] * 1
        done = False
        cum_rew = 0
        for action in actions:
            self.assertFalse(done)
            obs, reward, terminated, truncated, info = e.step(action)
            cum_rew += reward

        self.assertTrue(terminated)
        self.assertFalse(truncated)
        expected = 1.0 + len(actions) * e.neg_reward_scale
        self.assertAlmostEqual(cum_rew, expected)



from typing import Any, Dict, Optional, Tuple

import gymnasium as gym
import numpy as np
from gymnasium.core import ActType, ObsType

from popgym.core.deck import Deck
from popgym.core.env import POPGymEnv


class RepeatFirst(POPGymEnv):
    """A game where the agent must repeat the suit of the first card it saw

    Args:
        num_decks: The number of decks to cycle through, which determines
            episode length

    Returns:
        A gym environment
    """

    def __init__(self, num_decks=1):
        self.deck = Deck(num_decks)
        self.deck.add_players("player")
        self.max_episode_length = self.deck.num_cards - 1
        self.action_space = self.deck.get_obs_space(["suits"])
        self.observation_space = self.action_space
        self.state_space = gym.spaces.Tuple(
            (gym.spaces.Discrete(4), gym.spaces.Box(0, 1, (4,)))
        )
        self.dealt_cards = np.zeros((4,), dtype=int)

    def get_state(self):
        dealt_cards = 1.0 - self.dealt_cards / (self.deck.num_cards / 4)
        dealt_cards = dealt_cards.astype(np.float32)
        return self.card.copy(), dealt_cards

    def step(self, action: ActType) -> Tuple[ObsType, float, bool, bool, dict]:
        reward_scale = 1 / (self.deck.num_cards - 1)
        if action == self.card:
            reward = reward_scale
        else:
            reward = -reward_scale

        truncated = len(self.deck) == 1

        self.deck.deal("player", 1)
        card = self.deck.show("player", ["suits_idx"])[0, -1]
        self.dealt_cards[card] += 1
        obs = card.item()
        self.deck.discard_all()

        info: dict = {}

        return obs, reward, False, truncated, info

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict] = None,
    ) -> Tuple[gym.core.ObsType, Dict[str, Any]]:
        super().reset(seed=seed)
        self.deck.reset(rng=self.np_random)
        self.deck.deal("player", 1)
        self.dealt_cards[:] = 0
        self.card = self.deck.show("player", ["suits_idx"])[0, -1]
        self.dealt_cards[self.card] += 1
        obs = self.card.item()
        return obs, {}


class RepeatFirstEasy(RepeatFirst):
    def __init__(self):
        super().__init__(num_decks=1)


class RepeatFirstMedium(RepeatFirst):
    def __init__(self):
        super().__init__(num_decks=8)


class RepeatFirstHard(RepeatFirst):
    def __init__(self):
        super().__init__(num_decks=16)



from typing import Any, Dict, List, Tuple

import gym
from ray.rllib.utils.typing import ModelConfigDict, TensorType
from torch import nn

from popgym.baselines.ray_models.base_model import BaseModel


class MLP(BaseModel):
    """A good old MLP that has no memory whatsoever. Useful to see if your
    memory model is actually using its memory."""

    MODEL_CONFIG: Dict[str, Any] = {
        "embedding": "sine",
    }

    def __init__(
        self,
        obs_space: gym.spaces.Space,
        action_space: gym.spaces.Space,
        num_outputs: int,
        model_config: ModelConfigDict,
        name: str,
        **custom_model_kwargs,
    ):
        super().__init__(obs_space, action_space, num_outputs, model_config, name)
        self.core = nn.Sequential(
            nn.Linear(self.cfg["preprocessor_output_size"], self.cfg["hidden_size"]),
            nn.LeakyReLU(),
            nn.Linear(self.cfg["hidden_size"], self.cfg["hidden_size"]),
            nn.LeakyReLU(),
        )

    def initial_state(self) -> List[TensorType]:
        return []

    def forward_memory(
        self,
        z: TensorType,
        state: List[TensorType],
        t_starts: TensorType,
        seq_lens: TensorType,
    ) -> Tuple[TensorType, List[TensorType]]:

        z = self.core(z)

        # State expected to be list
        return z, []


class BasicMLP(MLP):
    MODEL_CONFIG = {
        "embedding": None,
    }



from typing import Optional

import gymnasium as gym
from gymnasium import spaces
from gymnasium.core import ObsType

from popgym.core.env import POPGymEnv


class POPGymWrapper(gym.Wrapper, POPGymEnv):
    def __init__(self, env: POPGymEnv):
        super().__init__(env)
        assert isinstance(env, POPGymEnv)
        self._state_space: Optional[spaces.Space] = None

    @property
    def state_space(self) -> spaces.Space:
        """Returns the observation space of the environment."""
        if self._state_space is None:
            return self.env.state_space
        return self._state_space

    @state_space.setter
    def state_space(self, space: spaces.Space):
        self._state_space = space

    def get_state(self) -> ObsType:
        return self.env.get_state()



import random
from enum import IntEnum
from typing import Optional
from warnings import warn

import gymnasium as gym
import numpy as np
from mazelib import Maze
from mazelib.generate.HuntAndKill import HuntAndKill

from popgym.core.env import POPGymEnv


class Explored(IntEnum):
    NO = 0
    YES = 1


class Cell(IntEnum):
    FREE = 0
    OBSTACLE = 1
    START = 2
    GOAL = 3
    HIDDEN = 4


class Actions(IntEnum):
    LEFT = 0
    RIGHT = 1
    UP = 2
    DOWN = 3
    NONE = 4


class MazeEnv(POPGymEnv):
    """A base class for maze-based environments.

    Args:
        maze_dims: The width and height of the maze in blocks
        episode_length: The maximum length of an episode

    Returns:
        A gym environment
    """

    obs_requires_prev_action = True

    def __init__(self, maze_dims=(10, 10), episode_length=1024):
        assert maze_dims[0] % 2 == 0 and maze_dims[1] % 2 == 0, "Maze dims must be even"
        self.maze_dims = [m // 2 for m in maze_dims]
        self.max_episode_length = episode_length
        # This could be MultiBinary but that uses int8 dtype
        # which causes RLlib obs preprocessor to crash
        self.observation_space = gym.spaces.MultiDiscrete(
            np.array(9 * [len(Cell)] + [len(Actions)], dtype=np.int32)
            # np.full(9, len(Cell), dtype=np.int32)
        )
        self.action_space = gym.spaces.Discrete(len(Actions) - 1)
        self.state_space = gym.spaces.MultiDiscrete(
            np.full([m + 1 for m in maze_dims], len(Cell))
        )
        warn(
            "Maze environments have been shown to be solvable without the use "
            "of memory (e.g. MLP outperforms LSTM). We suggest not using them "
            "as POMDP benchmarks.",
            DeprecationWarning,
        )

    def get_obs(self, action):
        view = self.local_view(self.maze.grid).reshape(9).astype(np.int32)
        obs = np.concatenate([view, np.array([action])], dtype=np.int32)
        return obs

    def move(self, action):
        y, x = self.position
        act_map = {
            0: (y, x - 1),  # left
            1: (y, x + 1),  # right
            2: (y - 1, x),  # up
            3: (y + 1, x),  # down
        }
        new_pos = act_map[action]
        new_y = min(max(new_pos[0], 0), 2 * self.maze_dims[0])
        new_x = min(max(new_pos[1], 0), 2 * self.maze_dims[1])
        new_pos = (new_y, new_x)

        if self.maze.grid[new_pos] == 1:
            # can't go thru obstacle
            return self.position

        return new_pos

    def local_view(self, grid):
        y, x = self.position
        x_start = max(x - 1, 0)
        x_end = min(x + 1, self.maze.grid.shape[0])
        # Y goes from top to bottom
        y_start = max(y - 1, 0)
        y_end = min(y + 1, self.maze.grid.shape[1])

        view = grid[y_start : y_end + 1, x_start : x_end + 1]
        y_pad, x_pad = 3 - view.shape[0], 3 - view.shape[1]
        padded = np.pad(view, [(0, y_pad), (0, x_pad)], constant_values=1)
        return padded

    def tostring(self, start=False, end=False, agent=False, visited=False):
        """Return a string representation of the maze.
        This can also display the maze entrances/solutions IF they already exist.
        Args:
            entrances (bool): Do you want to show the entrances of the maze?
            solutions (bool): Do you want to show the solution to the maze?
        Returns:
            str: string representation of the maze
        """
        # build the walls of the grid
        txt = []
        for i, row in enumerate(self.maze.grid):
            col = []
            for j, cell in enumerate(row):
                if visited and self.explored[i, j] == 1:
                    col += ["++"]
                elif cell == 0:
                    col += ["  "]
                elif cell == 1:
                    col += ["██"]
            txt.append("".join(col))

        # insert the start and end points
        if start and self.maze.start:
            if self.maze.start != self.position:
                r, c = self.maze.start
                txt[r] = txt[r][: 2 * c] + "🟢" + txt[r][2 * (c + 1) :]
        if end and self.maze.end:
            if self.maze.end != self.position:
                r, c = self.maze.end
                txt[r] = txt[r][: 2 * c] + "🛑" + txt[r][2 * (c + 1) :]
        if agent:
            r, c = self.position
            txt[r] = txt[r][: 2 * c] + "🚎" + txt[r][2 * (c + 1) :]

        return "\n".join(txt)

    def step(self, action):
        self.position = self.move(action)
        y, x = self.position
        self.explored[y, x] = Explored.YES
        self.curr_step += 1

    def get_state(self) -> gym.core.ObsType:
        state = self.maze.grid.copy()
        state[self.explored == Explored.NO] = Cell.HIDDEN
        return state

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict] = None,
    ):
        super().reset(seed=seed)
        if seed is not None:
            random.seed(seed)
            np.random.seed(seed)
        self.maze = Maze()
        self.maze.generator = HuntAndKill(*self.maze_dims)
        self.maze.generate()
        self.maze.generate_entrances(start_outer=True)
        self.position = self.maze.start
        y, x = self.position
        self.explored = np.full_like(self.maze.grid, Explored.NO)
        self.maze.grid[y, x] = Cell.START
        self.explored[y, x] = Explored.YES
        self.curr_step = 0



from typing import Any, Dict, Optional, Tuple

import gymnasium as gym
import numpy as np
from gymnasium.core import ActType, ObsType

from popgym.core.env import POPGymEnv


class MultiarmedBandit(POPGymEnv):
    """Multiarmed Bandits over an episode. Bandits are initialized that have some
    probability of positive reward and negative reward. The agent must sample
    and then exploit the bandits that pay best.

    Args:
        num_bandits: Number of individual bandits
        episode_length: Max episode length
    Returns:
        A gym env
    """

    obs_requires_prev_action = True

    def __init__(self, num_bandits=10, episode_length=200):
        self.num_bandits = num_bandits
        self.max_episode_length = episode_length
        self.observation_space = gym.spaces.Discrete(2)
        self.state_space = gym.spaces.Box(0, 1, (self.num_bandits,))
        self.action_space = gym.spaces.Discrete(num_bandits)
        self.num_steps = 0
        self.bandits = np.zeros((self.num_bandits,), dtype=np.float32)

    def get_state(self):
        return self.bandits.copy()

    def step(self, action: ActType) -> Tuple[ObsType, float, bool, bool, dict]:
        obs = int(self.np_random.random() < self.bandits[action])
        if obs:
            reward = 1 / self.max_episode_length
        else:
            reward = -1 / self.max_episode_length
        truncated = self.num_steps >= self.max_episode_length - 1
        self.num_steps += 1

        return obs, reward, False, truncated, self.info.copy()

    def reset(
        self,
        *,
        seed: Optional[int] = None,
        options: Optional[dict] = None,
    ) -> Tuple[gym.core.ObsType, Dict[str, Any]]:

        super(MultiarmedBandit, self).reset(seed=seed)
        self.num_steps = 0
        self.bandits = self.np_random.random(self.num_bandits, dtype=np.float32)
        self.info = {"bandits": self.bandits.copy()}
        obs = 0
        return obs, self.info


class MultiarmedBanditEasy(MultiarmedBandit):
    def __init__(self):
        super().__init__(num_bandits=10, episode_length=200)


class MultiarmedBanditMedium(MultiarmedBandit):
    def __init__(self):
        super().__init__(num_bandits=20, episode_length=400)


class MultiarmedBanditHard(MultiarmedBandit):
    def __init__(self):
        super().__init__(num_bandits=30, episode_length=600)



from typing import List, Tuple

import gym
import torch
from ray.rllib.utils.typing import ModelConfigDict, TensorType
from torch import nn

from popgym.baselines.ray_models.base_model import BaseModel


class Framestack(BaseModel):
    r"""Concatenate sequential frames along the feature dimension,
    pushing them through an MLP. From

    @article{mnih_human-level_2015,
        title = {Human-level control through deep reinforcement learning},
        volume = {518},
        number = {7540},
        journal = {nature},
        author = {
            Mnih, Volodymyr and Kavukcuoglu, Koray and Silver, David and Rusu,
            Andrei A and Veness, Joel and Bellemare, Marc G and Graves,
            Alex and Riedmiller, Martin and Fidjeland, Andreas K and Ostrovski,
            Georg and {others}
        },
        year = {2015},
        note = {Publisher: Nature Publishing Group},
        pages = {529--533},
    }
    """

    MODEL_CONFIG = {
        # Number of consecutive frames to stack into a single input
        "stack_size": 4,
        "embedding": "sine",
    }

    def __init__(
        self,
        obs_space: gym.spaces.Space,
        action_space: gym.spaces.Space,
        num_outputs: int,
        model_config: ModelConfigDict,
        name: str,
        **custom_model_kwargs,
    ):
        super().__init__(obs_space, action_space, num_outputs, model_config, name)
        # Compress before passing to framestack so we are not "cheating"
        # by using a larger hidden state than other models
        self.compress = nn.Linear(
            self.cfg["preprocessor_output_size"],
            self.cfg["hidden_size"] // self.cfg["stack_size"],
        )
        self.core = nn.Sequential(
            nn.Linear(
                self.cfg["hidden_size"],
                self.cfg["hidden_size"],
            ),
            nn.LeakyReLU(),
            nn.Linear(self.cfg["hidden_size"], self.cfg["hidden_size"]),
            nn.LeakyReLU(),
        )

    def initial_state(self) -> List[TensorType]:
        return [
            torch.zeros(
                self.cfg["stack_size"] - 1, self.cfg["preprocessor_output_size"]
            ),
        ]

    def forward_memory(
        self,
        z: TensorType,
        state: List[TensorType],
        t_starts: TensorType,
        seq_lens: TensorType,
    ) -> Tuple[TensorType, List[TensorType]]:

        assert self.cfg["hidden_size"] % self.cfg["stack_size"] == 0, (
            f"Hidden size {self.cfg['hidden_size']} must be divisible by stack size"
            f" {self.cfg['stack_size']}"
        )
        B, T, F = z.shape
        # [B, T + stack_size - 1, F]
        time_padded = torch.cat([state[0], z], dim=1)
        # [B, T + stack_size - 1, f, k * T]
        unfolded = time_padded.unfold(1, self.cfg["stack_size"], 1)
        # [B * T, stack_size, k]
        stacked = unfolded.permute(0, 1, 3, 2).reshape(
            B * T, self.cfg["stack_size"], z.shape[-1]
        )
        # Add extra dim so compression is applied frame-wise, then fold
        # new dim back into feature dim
        compressed = self.compress(
            stacked.reshape(
                B, T, self.cfg["stack_size"], self.cfg["preprocessor_output_size"]
            )
        ).reshape(B, T, self.cfg["hidden_size"])
        z = self.core(compressed)
        # Output last stack_size - 1 from stacked at time T
        state = stacked.reshape(
            B, T, self.cfg["stack_size"], self.cfg["preprocessor_output_size"]
        )[:, -1, 1:]

        # State expected to be list
        return z, [state]



from popgym.envs.repeat_previous import RepeatPrevious
from tests.base_env_test import AbstractTest


class TestRepeatPrevious(AbstractTest.POPGymTest):
    def setUp(self) -> None:
        self.env = RepeatPrevious()

    def test_k(self):
        e = RepeatPrevious(k=2)
        obs0, _ = e.reset()
        obs0, _ = e.reset()
        obs1, rew1, te1, tr1, info1 = e.step(0)
        self.assertFalse(te1 or tr1)
        obs2, rew2, te2, tr2, info2 = e.step(obs0)
        self.assertFalse(te2 or tr2)
        obs3, rew3, te3, tr3, info3 = e.step(obs1)
        self.assertFalse(te3 or tr3)

    def test_full(self):
        e = RepeatPrevious(k=2)
        obs1, _ = e.reset()
        cum_rew = 0
        obs2, reward, terminated, truncated, info = e.step(0)
        cum_rew += reward

        all_obs = [obs1, obs2]

        iters = 2
        while not (terminated or truncated):
            a = all_obs[-2]
            obs, reward, terminated, truncated, info = e.step(a)
            all_obs.append(obs)
            cum_rew += reward
            iters += 1

        self.assertAlmostEqual(cum_rew, 1.0)

        self.assertEqual(iters, 52)

    def test_full_16(self):
        e = RepeatPrevious(k=16)
        o, _ = e.reset()
        all_obs = [o]
        iters = 1
        cum_rew = 0
        while iters < 16:
            obs, reward, terminated, truncated, info = e.step(0)
            self.assertFalse(terminated or truncated)
            self.assertEqual(reward, 0)
            cum_rew += reward
            all_obs.append(obs)
            iters += 1

        while iters < 52:
            a = all_obs[-16]
            self.assertFalse(terminated or truncated)
            obs, reward, terminated, truncated, info = e.step(a)
            cum_rew += reward
            all_obs.append(obs)
            iters += 1

        self.assertTrue(truncated)
        self.assertFalse(terminated)
        self.assertAlmostEqual(cum_rew, 1.0)
